import asyncio
from typing import Dict, List, Union, Any

from ..Common import fetch_api_data, cache_manager, session_manager


class MarketUser:
    """
    Represents a user on warframe.market
    """
    base_api_url: str = "https://api.warframe.market/v1"  # Base URL for warframe.market API
    base_url: str = "https://warframe.market/profile"  # Base URL for warframe.market profiles
    asset_url: str = "https://warframe.market/static/assets"  # Base URL for warframe.market assets

    def __init__(self, database: "MarketDatabase", user_id: str, username: str):
        """
        Initializes a MarketUser object.
        :param database: database object
        :param user_id: the user's warframe.market id
        :param username: the user's warframe.market username
        """
        self.database = database
        self.user_id = user_id
        self.username = username
        self.profile_url: str = f"{MarketUser.base_url}/{self.username}"
        self.last_seen = None
        self.avatar = None
        self.avatar_url = None
        self.locale = None
        self.background = None
        self.about = None
        self.reputation = None
        self.platform = None
        self.banned = None
        self.status = None
        self.region = None
        self.orders: Dict[str, List[Dict[str, Union[str, int]]]] = {'buy': [], 'sell': []}
        self.reviews: List[str] = []

    @classmethod
    async def create(cls, database: "MarketDatabase", user_id: str, username: str,
                     fetch_user_data: bool = True, fetch_orders: bool = True, fetch_reviews: bool = True):
        """
        Creates a MarketUser object.
        :param database: database object
        :param user_id: the user's warframe.market id
        :param username: the user's warframe.market username
        :param fetch_user_data: whether or not to fetch the user's data from warframe.market
        :param fetch_orders: whether or not to fetch the user's orders from warframe.market
        :param fetch_reviews: whether or not to fetch the user's reviews from warframe.market
        :return: MarketUser object with the specified data
        """
        obj = cls(database, user_id, username)

        tasks = []
        if fetch_user_data:
            profile = await MarketUser.fetch_user_data(username)
            if profile is not None:
                obj.set_user_data(profile)

        if fetch_orders:
            tasks.append(obj.fetch_orders())

        if fetch_reviews:
            tasks.append(obj.fetch_reviews())

        await asyncio.gather(*tasks)

        return obj

    @staticmethod
    async def fetch_user_data(username) -> Union[None, Dict]:
        """
        Fetches a user's data from warframe.market
        :param username: the user's warframe.market username, case-sensitive
        :return: the user's data
        """
        async with cache_manager() as cache, session_manager() as session:
            user_data = await fetch_api_data(session=session,
                                             cache=cache,
                                             url=f"{MarketUser.base_api_url}/profile/{username}")
        if user_data is None:
            return

        # Load the user profile
        try:
            profile = user_data['payload']['profile']
        except KeyError:
            return

        return profile

    def set_user_data(self, profile: Dict[str, Any]) -> None:
        """
        Sets the user's data based on the values returned from the API.
        :param profile: the user's profile data
        :return: None
        """
        for key, value in profile.items():
            if hasattr(self, key):
                setattr(self, key, value)

        if self.avatar is not None:
            self.avatar_url = f"{MarketUser.asset_url}/{self.avatar}"

    def parse_orders(self, orders: List[Dict[str, Any]]) -> None:
        """
        Parses the user's orders.
        :param orders: the user's orders from the API
        :return: None
        """
        self.orders: Dict[str, List[Dict[str, Union[str, int]]]] = {'buy': [], 'sell': []}

        for order_type in ['sell_orders', 'buy_orders']:
            for order in orders[order_type]:
                parsed_order = {
                    'item': order['item']['en']['item_name'],
                    'item_url_name': order['item']['url_name'],
                    'item_id': order['item']['id'],
                    'last_update': order['last_update'],
                    'quantity': order['quantity'],
                    'price': order['platinum'],
                }

                if 'subtype' in order:
                    parsed_order['subtype'] = order['subtype']

                if 'mod_rank' in order:
                    parsed_order['subtype'] = f"R{order['mod_rank']}"

                self.orders[order_type.split('_')[0]].append(parsed_order)

    def parse_reviews(self, reviews: List[Dict[str, Any]]) -> None:
        """
        Parses the user's reviews.
        :param reviews: the user's reviews from the API
        :return: None
        """
        for review in reviews:
            parsed_review = {
                'user': review['user_from']['ingame_name'],
                'user_id': review['user_from']['id'],
                'user_avatar': review['user_from']['avatar'],
                'user_region': review['user_from']['region'],
                'text': review['text'],
                'date': review['date'],
            }

            if parsed_review not in self.reviews:
                self.reviews.append(parsed_review)

    async def fetch_orders(self) -> None:
        """
        Fetches the user's orders from warframe.market
        :return: None
        """
        async with cache_manager() as cache, session_manager() as session:
            orders = await fetch_api_data(cache=cache,
                                          session=session,
                                          url=f"{self.base_api_url}/profile/{self.username}/orders",
                                          expiration=60)

        if orders is None:
            return

        self.parse_orders(orders['payload'])

    async def fetch_reviews(self, page_num: str = '1') -> None:
        """
        Fetches the user's reviews from warframe.market
        :param page_num: the page number to fetch
        :return: None
        """
        async with cache_manager() as cache, session_manager() as session:
            reviews = await fetch_api_data(cache=cache,
                                           session=session,
                                           url=f"{self.base_api_url}/profile/{self.username}/reviews/{page_num}",
                                           expiration=60)

        if reviews is None:
            return

        self.parse_reviews(reviews['payload']['reviews'])



import asyncio
import json
import os
from typing import Dict, List, Tuple, Any

import aiohttp
import redis
from aiohttp import ClientResponseError
from bs4 import BeautifulSoup

from market_engine.Common import fetch_api_data, logger, config, fix_names_and_add_ids, get_platform_path, \
    get_statistic_path

RELICS_RUN_BASE_URL = "https://relics.run"  # Base URL for relics.run
RELICS_RUN_HISTORY_URL = f"{RELICS_RUN_BASE_URL}/history"  # URL for fetching statistic history


async def fetch_statistics_from_relics_run(cache: redis.Redis,
                                           session: aiohttp.ClientSession,
                                           item_ids: Dict[str, str],
                                           date_list: List[str],
                                           platform: str = 'pc') -> None:
    """
    Fetches item statistics from relics.run, and saves them to the output directory as specified in the config.
    :param cache: redis cache
    :param session: aiohttp session
    :param item_ids: dictionary of item names to item ids
    :param date_list: list of dates to fetch
    :param platform: platform to fetch statistics for
    return None
    """

    async def fetch_data(date):
        url = f"https://relics.run/history/{get_platform_path(platform)}{date}"

        try:
            data = await fetch_api_data(session=session, url=url)
        except ClientResponseError:
            logger.error(f"Failed to fetch data for {url}")
            return

        await fix_names_and_add_ids(data, translation_dict, item_ids)

        with open(os.path.join(config['output_dir'], date), 'w') as f:
            json.dump(data, f)

    translation_dict = await fetch_translation_dict_from_relics_run(cache, session)

    await asyncio.gather(*[fetch_data(date) for date in date_list])


async def get_all_saved_dates_from_relics_run(cache: redis.Redis,
                                              session: aiohttp.ClientSession,
                                              platform: str = 'pc') -> set:
    """
    Fetches the dates of all saved statistics from relics.run
    :param platform: platform to fetch statistics for
    :param cache: redis cache
    :param session: aiohttp session
    :return: set of saved dates
    """
    data = await fetch_api_data(cache=cache,
                                session=session,
                                url=RELICS_RUN_HISTORY_URL + get_platform_path(platform),
                                return_type='text')

    # Parses the HTML and finds all links to JSON files
    soup = BeautifulSoup(data, 'html.parser')

    urls = set()
    for link_obj in soup.find_all('a'):
        link = link_obj.get('href')
        if link.endswith('json'):
            urls.add(link)

    return urls


def get_saved_data(platform: str = 'pc') -> set:
    """
    Gets the names of all saved statistics from the output directory
    :param platform: platform for which to get saved statistics
    :return: set of saved statistics
    """
    saved_data = set()
    if not os.path.exists(config['output_dir']):
        os.makedirs(config['output_dir'])

    for file in os.listdir(get_statistic_path(platform)):
        if file.endswith(".json"):
            saved_data.add(file)

    return saved_data


async def get_dates_to_fetch(cache: redis.Redis,
                             session: aiohttp.ClientSession) -> set:
    """
    Gets the dates for which statistics need to be fetched
    :param cache: redis cache
    :param session: aiohttp session
    :return: set of dates to fetch
    """
    date_list = await get_all_saved_dates_from_relics_run(cache, session)  # Get all possible dates
    saved_data = get_saved_data()  # Get the dates for which statistics have already been fetched
    date_list = date_list - saved_data  # Remove the dates for which statistics have already been fetched

    return date_list


async def fetch_item_ids_from_relics_run(cache: redis.Redis,
                                         session: aiohttp.ClientSession) -> Dict[str, str]:
    """
    Fetches item ids from relics.run
    :param cache: redis cache
    :param session: aiohttp session
    :return: dictionary of item names to item ids
    """
    url = f"{RELICS_RUN_BASE_URL}/market_data/item_ids.json"
    return await fetch_api_data(cache=cache, session=session, url=url)


async def fetch_item_info_from_relics_run(cache: redis.Redis,
                                          session: aiohttp.ClientSession) -> Dict[str, str]:
    """
    Fetches item info from relics.run
    :param cache: redis cache
    :param session: aiohttp session
    :return: dictionary of item names to item info
    """
    url = f"{RELICS_RUN_BASE_URL}/market_data/item_info.json"
    return await fetch_api_data(cache=cache, session=session, url=url)


async def fetch_items_from_relics_run(cache: redis.Redis,
                                      session: aiohttp.ClientSession) -> List[Dict[str, str]]:
    """
    Fetches list of warframe.market items from relics.run
    :param cache: redis cache
    :param session: aiohttp session
    :return: list of warframe.market items
    """
    url = f"{RELICS_RUN_BASE_URL}/market_data/items.json"
    return await fetch_api_data(cache=cache, session=session, url=url)


async def fetch_translation_dict_from_relics_run(cache: redis.Redis,
                                                 session: aiohttp.ClientSession) -> Dict[str, str]:
    """
    Fetches translation dictionary from relics.run
    :param cache: redis cache
    :param session: aiohttp session
    :return: Dictionary translating old/changed item names to corrected versions
    """
    url = f"{RELICS_RUN_BASE_URL}/market_data/translation_dict.json"
    return await fetch_api_data(cache=cache, session=session, url=url)


async def fetch_item_data_from_relics_run(cache: redis.Redis,
                                          session: aiohttp.ClientSession) -> list[Any]:
    """
    Fetches all item data from relics.run
    :param cache: redis cache
    :param session: aiohttp session
    :return: list of items, item_ids, item_info, and translation_dict
    """
    tasks = [fetch_items_from_relics_run(cache, session),
             fetch_item_ids_from_relics_run(cache, session),
             fetch_item_info_from_relics_run(cache, session),
             fetch_translation_dict_from_relics_run(cache, session)]

    return await asyncio.gather(*tasks)
